// Punch the Monkey
// a simple game of senseless violence

// please enjoy the fact that this game doesn't require 4 files to work
// run with "g++ posix_monkey_puncher.cc -o puncher -l pthread && ./puncher"

// adam@adamsmith.as

#include <iostream>
#include <cstdlib>
#include <pthread.h>
using namespace std;


class Game {
private:
    bool punchable;    
    int score;
    pthread_mutex_t mutex;
public:
    Game() :punchable(false), score(0) {};
    void run();
    void punch();
};

class KeyboardInput {
private:
    static void* worker_func (void* data); 
    pthread_t worker;    
public:    
    KeyboardInput(Game *g) {
        pthread_create(&worker, NULL, worker_func, static_cast<void*>(g));
    }

    ~KeyboardInput() {
        pthread_cancel(worker);
    }

};
    
void Game::run() {
        
    cout << "***************************************************" << endl;
    cout << "***** P O S I X   M o n k e y   P u n c h e r *****" << endl;
    cout << "***************************************************" << endl;
    cout << endl;
    cout << "A pthread demo program illustrating an" << endl;
    cout << "input manager object using a blocking " << endl;
    cout << "operation in a loop." << endl;
    cout << endl;
    cout << "Instructions: Punch the monkey only when" << endl;
    cout << "it is punchable to maximize your score." << endl;
    cout << "Hit ENTER to punch.  Punch when you are" << endl;
    cout << "ready to begin the game." << endl;  
    cout << endl;
    
    cin.ignore(INT_MAX, '\n');
    pthread_mutex_init(&mutex, NULL);
    KeyboardInput kbd(this); 
    
    
    while(true) {

        pthread_mutex_lock(&mutex);
        // first order markov monkey model
        if(punchable) {
            if(!(rand()%3)) { punchable = false; }
        } else {
            if(!(rand()%8)) { punchable = true; }
        }
        const char* str = punchable?"quite ":"hardly";
        cout << "The monkey is " << str << " punchable. SCORE: " << score << endl;
        pthread_mutex_unlock(&mutex);

        usleep(100000);

        if(score>0) {
            break;
        }
    }
    
    cout << "Oh, come on. Lay off the monkey will ya?" << endl;

    // kbd's dtor is called right here before it goe out of scope 
} // Game::run
        
void Game::punch() {
    pthread_mutex_lock(&mutex);
    if(punchable) {
        score++;

        cout << endl;
        cout << "***********" << endl;
        cout << "** SMACK ** You punched the monkey right in the kisser!" << endl;
        cout << "***********" << endl;
        cout << endl;
    } else {
        score--;
        
        cout << endl;
        cout << "..........." << endl;
        cout << ".. SWISH .. You merely waved at the monkey." << endl;
        cout << "..........." << endl;
        cout << endl;
    }
    pthread_mutex_unlock(&mutex);
} // Game::punch

void* KeyboardInput::worker_func (void* data) {
    Game* g = static_cast<Game*>(data);

    while(true) {
        cin.ignore(INT_MAX, '\n');
        g->punch();
    }

    return NULL;
} // KeyboardInput::worker_func

int main(int argc, char** argv) {
    Game g;
    g.run();
    return 0;
} // main
